import uuid

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from .settings import DJANGO_SMS_TOOLKIT_SETTINGS


class TwilioMessage(models.Model):

    # Outgoing Statuses
    STATUS_ACCEPTED = "accepted"

    # This is a status that is not exist in the Twilio.
    # It will be used for the messages that is not accepted by the twilio.
    # For example in a case where phone number is not valid.
    STATUS_REJECTED = "rejected"

    STATUS_QUEUED = "queued"
    STATUS_SENT = "sent"
    STATUS_DELIVERED = "delivered"
    STATUS_UNDELIVERED = "undelivered"
    STATUS_FAILED = "failed"

    # Incoming Statuses
    STATUS_RECEIVED = "received"

    STATUSES = (
        (STATUS_REJECTED, STATUS_REJECTED),
        (STATUS_ACCEPTED, STATUS_ACCEPTED),
        (STATUS_QUEUED, STATUS_QUEUED),
        (STATUS_SENT, STATUS_SENT),
        (STATUS_DELIVERED, STATUS_DELIVERED),
        (STATUS_UNDELIVERED, STATUS_UNDELIVERED),
        (STATUS_FAILED, STATUS_FAILED),
        (STATUS_RECEIVED, STATUS_RECEIVED),
    )

    TYPE_INCOMING = "incoming"
    TYPE_OUTGOING = "outgoing"

    TYPES = (
        (TYPE_INCOMING, TYPE_INCOMING),
        (TYPE_OUTGOING, TYPE_OUTGOING),
    )

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    recipient = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="twilio_messages", on_delete=models.SET_NULL, null=True, blank=True)
    message_sid = models.CharField(max_length=512, null=True, blank=True, unique=True)
    type = models.CharField(choices=TYPES, max_length=32)
    status = models.CharField(choices=STATUSES, max_length=32, blank=True)
    from_number = models.CharField(max_length=255)
    to_number = models.CharField(max_length=255)

    # The message body aimed to be send. If the message trimming is active and
    # message is too long, `body` will store the trimmed version of the message.
    # For example:
    # original_body = ".....and say hello lorem sit...."
    # body = ".......and s"
    original_body = models.TextField()
    body = models.TextField()

    # TwilioRestException message if exists.
    exception_message = models.TextField()

    creation_datetime = models.DateTimeField(default=timezone.now)
    update_datetime = models.DateTimeField(auto_now=True)

    def __str__(self):
        return "Twilio message from {from_number} to {to_number}".format(from_number=self.from_number, to_number=self.to_number)

    class Meta:
        verbose_name = _('Twilio Message')
        verbose_name_plural = _('Twilio Messages')
        ordering = ["-creation_datetime"]

    def save(self, *args, **kwargs):
        self.full_clean()
        super().save(*args, **kwargs)

    def clean(self):
        super(TwilioMessage, self).clean()

        if self.type == self.TYPE_INCOMING and self.status and self.status != self.STATUS_RECEIVED:
            raise ValidationError("Incoming message status is not valid.")

        if self.type == self.TYPE_OUTGOING and self.status == self.STATUS_RECEIVED:
            raise ValidationError("Outgoing message status is not valid.")

        if not self._state.adding and self.type == self.TYPE_OUTGOING:
            # https://support.twilio.com/hc/en-us/articles/223134347-What-are-the-Possible-SMS-and-MMS-Message-Statuses-and-What-do-They-Mean-
            ongoing_statuses = [self.STATUS_ACCEPTED, self.STATUS_QUEUED, self.STATUS_SENT]
            finalized_statuses = [self.STATUS_DELIVERED, self.STATUS_UNDELIVERED, self.STATUS_FAILED]

            # https://docs.djangoproject.com/en/2.2/ref/models/instances/#refreshing-objects-from-database
            new_status = self.status
            del self.status
            # Fetch status from database.
            old_status = self.status

            if old_status in finalized_statuses and new_status in ongoing_statuses:
                raise ValidationError(
                    "Cannot revert status of outgoing message ({id}) from {old_status} to {new_status}".format(
                        id=self.id,
                        old_status=old_status,
                        new_status=new_status,
                    )
                )
            else:
                self.status = new_status


class TwilioMessageMixin(object):

    def get_phone_number(self):
        return self.phone_number

    def send_sms(self, body, from_number=None):
        from .tasks import send_sms

        if not DJANGO_SMS_TOOLKIT_SETTINGS["SEND_SMS"]:
            return 'Cannot send SMS because of environment settings.'

        from_number = from_number or DJANGO_SMS_TOOLKIT_SETTINGS["DEFAULT_FROM_NUMBER"]
        assert from_number

        if isinstance(self, get_user_model()):
            recipient_id = self.pk
        else:
            recipient_id = None

        send_sms.delay(from_number, self.get_phone_number(), body, recipient_id=recipient_id)



# Generated by Django 2.1.7 on 2019-09-04 12:15

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='TwilioMessage',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('message_sid', models.CharField(blank=True, max_length=512, null=True, unique=True)),
                ('type', models.CharField(choices=[('incoming', 'incoming'), ('outgoing', 'outgoing')], max_length=32)),
                ('status', models.CharField(blank=True, choices=[('accepted', 'Accepted'), ('queued', 'Queued'), ('sent', 'Sent'), ('delivered', 'Delivered'), ('undelivered', 'Undelivered'), ('failed', 'Failed'), ('received', 'Received')], max_length=32)),
                ('from_number', models.CharField(max_length=255)),
                ('to_number', models.CharField(max_length=255)),
                ('original_body', models.TextField()),
                ('body', models.TextField()),
                ('exception_message', models.TextField()),
                ('creation_datetime', models.DateTimeField(default=django.utils.timezone.now)),
                ('update_datetime', models.DateTimeField(auto_now=True)),
                ('recipient', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='twilio_messages', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Twilio Message',
                'verbose_name_plural': 'Twilio Messages',
            },
        ),
    ]



from celery import shared_task
from django.urls import reverse
from twilio.base.exceptions import TwilioRestException
from twilio.rest import Client

from .models import TwilioMessage
from .settings import DJANGO_SMS_TOOLKIT_SETTINGS


@shared_task(retry_backoff=30, retry_kwargs={'max_retries': 3})
def send_sms(from_number, to_number, body, recipient_id=None):
    if not DJANGO_SMS_TOOLKIT_SETTINGS["SEND_SMS"]:
        return 'Cannot send SMS because of environment settings.'

    # Twilio does not accept body larger than 1600 characters. https://www.twilio.com/docs/api/errors/21617
    # If the body is larger than 1600, Send the first 1600 chars.
    original_body = body

    if DJANGO_SMS_TOOLKIT_SETTINGS["TWILIO"]["TRIM_LONG_BODY"]:
        body = body[0:1600]

    # Create twilio message before calling twilio.
    # If twilio message object created after message is sent to Twilio, race condition occurs.
    # Some times Twilio calls status callback before the transaction is completed and message is saved.
    twilio_message = TwilioMessage.objects.create(
        recipient_id=recipient_id,
        type=TwilioMessage.TYPE_OUTGOING,
        from_number=from_number,
        to_number=to_number,
        original_body=original_body,
        body=body
    )

    client = Client(DJANGO_SMS_TOOLKIT_SETTINGS["TWILIO"]["ACCOUNT_SID"], DJANGO_SMS_TOOLKIT_SETTINGS["TWILIO"]["AUTH_TOKEN"])

    try:
        message = client.messages.create(
            from_=from_number,
            to=to_number,
            body=body,
            status_callback=DJANGO_SMS_TOOLKIT_SETTINGS["TWILIO"]["STATUS_CALLBACK_BASE_URL"].strip("/") + reverse("twilio-status-callback", args=[twilio_message.id])
        )
        # NOTE: Status will be saved by status callback view.
        twilio_message.message_sid = message.sid
        twilio_message.save(update_fields=["message_sid"])
    except TwilioRestException as error:
        twilio_message.status = TwilioMessage.STATUS_REJECTED
        twilio_message.exception_message = error.__repr__()
        twilio_message.save()

        if error.code in (21211, 21614):  # Invalid Number
            return "Phone number is invalid. Failed saliently."
        else:
            raise error



from functools import wraps

from django.conf import settings
from django.db import transaction
from django.http import HttpResponse, HttpResponseForbidden
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
from twilio.request_validator import RequestValidator
from .settings import DJANGO_SMS_TOOLKIT_SETTINGS

from .models import TwilioMessage


def validate_twilio_request(f):
    """
    Validates that incoming requests genuinely originated from Twilio
    Docs: https://www.twilio.com/docs/usage/tutorials/how-to-secure-your-django-project-by-validating-incoming-twilio-requests
    """
    @wraps(f)
    def decorated_function(request, *args, **kwargs):
        # Create an instance of the RequestValidator class
        validator = RequestValidator(DJANGO_SMS_TOOLKIT_SETTINGS["TWILIO"]["AUTH_TOKEN"])

        # Validate the request using its URL, POST data,
        # and X-TWILIO-SIGNATURE header
        request_valid = validator.validate(
            request.build_absolute_uri(),
            request.POST,
            request.META.get('HTTP_X_TWILIO_SIGNATURE', '')
        )

        # Continue processing the request if it's valid, return a 403 error if
        # it's not
        if request_valid or settings.DEBUG:
            return f(request, *args, **kwargs)
        else:
            return HttpResponseForbidden()
    return decorated_function


@require_POST
@csrf_exempt
@validate_twilio_request
def twilio_status_callback_view(request, message_pk):
    with transaction.atomic():
        message = TwilioMessage.objects.select_for_update().get(pk=message_pk)
        message.status = request.POST["SmsStatus"]
        message.save()

    return HttpResponse(status=200)



# Generated by Django 2.2.6 on 2019-10-28 13:53

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_sms_toolkit', '0002_alter_model_options'),
    ]

    operations = [
        migrations.AlterField(
            model_name='twiliomessage',
            name='recipient',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='twilio_messages', to=settings.AUTH_USER_MODEL),
        ),
    ]



# Generated by Django 2.1.7 on 2019-10-09 12:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_sms_toolkit', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='twiliomessage',
            options={'ordering': ['-creation_datetime'], 'verbose_name': 'Twilio Message', 'verbose_name_plural': 'Twilio Messages'},
        ),
        migrations.AlterField(
            model_name='twiliomessage',
            name='status',
            field=models.CharField(blank=True, choices=[('rejected', 'rejected'), ('accepted', 'accepted'), ('queued', 'queued'), ('sent', 'sent'), ('delivered', 'delivered'), ('undelivered', 'undelivered'), ('failed', 'failed'), ('received', 'received')], max_length=32),
        ),
    ]



from django.contrib import admin

from .models import TwilioMessage


class TwilioMessageAdmin(admin.ModelAdmin):

    search_fields = ["message_sid", "original_body", "from_number", "to_number"]
    list_filter = ["status", "type"]
    list_display = ["id", "message_sid", "recipient", "from_number", "to_number", "body", "type", "status"]

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False


admin.site.register(TwilioMessage, TwilioMessageAdmin)



from django.conf import settings

DJANGO_SMS_TOOLKIT_SETTINGS = getattr(settings, "DJANGO_SMS_TOOLKIT_SETTINGS", {})

DJANGO_SMS_TOOLKIT_SETTINGS.setdefault("SEND_SMS", True)
DJANGO_SMS_TOOLKIT_SETTINGS.setdefault("DEFAULT_FROM_NUMBER", "")
DJANGO_SMS_TOOLKIT_SETTINGS.setdefault("TWILIO", {})
DJANGO_SMS_TOOLKIT_SETTINGS["TWILIO"].setdefault("ACCOUNT_SID", "")
DJANGO_SMS_TOOLKIT_SETTINGS["TWILIO"].setdefault("AUTH_TOKEN", "")
DJANGO_SMS_TOOLKIT_SETTINGS["TWILIO"].setdefault("STATUS_CALLBACK_BASE_URL", "")
DJANGO_SMS_TOOLKIT_SETTINGS["TWILIO"].setdefault("TRIM_LONG_BODY", True)


if DJANGO_SMS_TOOLKIT_SETTINGS["SEND_SMS"]:
    required_twilio_fields = ["ACCOUNT_SID", "AUTH_TOKEN", "STATUS_CALLBACK_BASE_URL"]
    for field_name in required_twilio_fields:
        if not DJANGO_SMS_TOOLKIT_SETTINGS["TWILIO"].get(field_name):
            assert False, "DJANGO_SMS_TOOLKIT_SETTINGS['TWILIO']: {} is required.".format(field_name)



import os
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-sms-toolkit',
    version='0.1.5',
    packages=['django_sms_toolkit'],
    include_package_data=True,
    install_requires=['twilio >= 6.0.0'],
    license='Apache-2.0',
    description="It's a Django module to send and store SMS with Twilio.",
    long_description=README,
    long_description_content_type="text/markdown",
    author='hipo',
    author_email='pypi@hipolabs.com',
    url='https://github.com/Hipo/django-sms-toolkit',
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*",
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



from django.conf.urls import url
from .views import twilio_status_callback_view

urlpatterns = [
    url(r'^twilio-status-callback/(?P<message_pk>[0-9a-f-]+)/$', twilio_status_callback_view, name="twilio-status-callback"),
]
